package mygame;

import com.jme3.app.SimpleApplication;
import com.jme3.input.KeyInput;
import com.jme3.input.controls.AnalogListener;
import com.jme3.input.controls.KeyTrigger;
import com.jme3.light.DirectionalLight;
import com.jme3.material.Material;
import com.jme3.math.ColorRGBA;
import com.jme3.math.Vector2f;
import com.jme3.math.Vector3f;
import com.jme3.scene.Geometry;
import com.jme3.scene.Mesh;
import com.jme3.scene.Node;
import com.jme3.scene.VertexBuffer.Type;
import com.jme3.scene.shape.Box;
import com.jme3.util.BufferUtils;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import static java.lang.Math.signum;
import java.util.ArrayList;

public class Main extends SimpleApplication {

    Vector3f[] verticesBody, verticesHead, verticesLeftLeg, verticesLeftArm, verticesRightLeg, verticesRightArm;
    float[] normals;
    Vector2f[] texCoords;
    int[] indexes;
    Node robot, box1Node, box2Node;
    Vector3f max1, min1, max2, min2;
    Vector3f Direction;
    //To prevent infinite loops - if an intersection is not found in 20 rounds, consider there is no intersection.
    final int MaxIterations = 20;

    public static void main(String[] args) {
        Main app = new Main();
        //app.setDisplayFps(false);       // TODO: UNCOMMENT TO HIDE FPS INFO ON APPLICATION RUN
        app.setDisplayStatView(false);  // TODO: UNCOMMENT TO HIDE STATS VIEW INFO ON APPLICATION RUN
        app.setShowSettings(false);     // TODO: UNCOMMENT TO HIDE SETTINGS MENU ON STARTUP
        app.start();
    }

    @Override
    public void simpleInitApp() {
        initKeys();
        flyCam.setMoveSpeed(10);

        /**
         * BEGIN OF EXERCISE 1 (Step 1): complete the adding models to the scene
         * graph code in the following TODO's for Step 1.
         */
        Mesh meshBody = new Mesh();
        Mesh meshHead = new Mesh();
        Mesh meshLeftLeg = new Mesh();
        Mesh meshRightLeg = new Mesh();
        Mesh meshLeftArm = new Mesh();
        Mesh meshRightArm = new Mesh();

        parseFile("assets/Models/model.data");

        meshBody.setBuffer(Type.Position, 3, BufferUtils.createFloatBuffer(verticesBody));
        meshBody.setBuffer(Type.Normal, 3, BufferUtils.createFloatBuffer(normals));
        meshBody.setBuffer(Type.TexCoord, 2, BufferUtils.createFloatBuffer(texCoords));
        meshBody.setBuffer(Type.Index, 3, BufferUtils.createIntBuffer(indexes));
        meshBody.updateBound();

        // head
        // TODO: (Sub-Step 1.1) ADD YOUR CODE HERE!
        

        // left leg
        // TODO: (Sub-Step 1.2) ADD YOUR CODE HERE!
        

        // right leg
        // TODO: (Sub-Step 1.3) ADD YOUR CODE HERE!
        

        // left arm
        // TODO: (Sub-Step 1.4) ADD YOUR CODE HERE!
        

        // right arm
        // TODO: (Sub-Step 1.5) ADD YOUR CODE HERE!
        

        // END OF EXERCISE 1 (Step 1). ----------------------------------------- 

        Geometry geoBody = new Geometry("ColoredMesh1", meshBody);
        Geometry geoHead = new Geometry("ColoredMesh2", meshHead);
        Geometry geoLeftLeg = new Geometry("ColoredMesh3", meshLeftLeg);
        Geometry geoRightLeg = new Geometry("ColoredMesh4", meshRightLeg);
        Geometry geoLeftArm = new Geometry("ColoredMesh5", meshLeftArm);
        Geometry geoRightArm = new Geometry("ColoredMesh6", meshRightArm);

        Material mat = new Material(assetManager, "Common/MatDefs/Misc/Unshaded.j3md");
        mat.setColor("Color", ColorRGBA.Blue);

        geoBody.setMaterial(mat);
        geoHead.setMaterial(mat);
        geoLeftLeg.setMaterial(mat);
        geoRightLeg.setMaterial(mat);
        geoLeftArm.setMaterial(mat);
        geoRightArm.setMaterial(mat);

        robot = new Node("Robot");
        robot.attachChild(geoBody);
        robot.attachChild(geoHead);
        robot.attachChild(geoLeftLeg);
        robot.attachChild(geoRightLeg);
        robot.attachChild(geoLeftArm);
        robot.attachChild(geoRightArm);

        rootNode.attachChild(robot);

        max1 = new Vector3f(.26f, .335f, 8); // the maximum point that defines the box 1.
        min1 = new Vector3f(-.26f, -.355f, 8);   // the minimum point that defines the box 1.

        // AABB 1
        Material box1Mat = new Material(assetManager,
                "Common/MatDefs/Misc/Unshaded.j3md");
        box1Mat.setColor("Color", ColorRGBA.Red);
        box1Mat.getAdditionalRenderState().setWireframe(true);
        Box box1 = new Box(min1, max1);
        Geometry box1Geom = new Geometry("box1", box1);
        box1Geom.setMaterial(box1Mat);
        box1Node = new Node("box1Node");
        box1Node.attachChild(box1Geom);
        rootNode.attachChild(box1Node);

        max2 = new Vector3f(1, .25f, 8); // the maximum point that defines the box 2.
        min2 = new Vector3f(.5f, -.25f, 8);   // the minimum point that defines the box 2.

        // AABB 2
        /**
         * BEGIN OF EXERCISE 2 (Step 2): complete the adding a second axis
         * aligned bounding box (AABB) to the scene graph code in the following
         * TODO's for Step 2.
         */
        // TODO: (Sub-Step 2.1) ADD YOUR CODE HERE!
        

        // END OF EXERCISE 2 (Step 2). ----------------------------------------- 

        /* adding lights to the scene graph */
        // add a light at (10, 20, 10) pointing to (-X. -Y, -Z)
        DirectionalLight sun2 = new DirectionalLight();
        sun2.setDirection(new Vector3f(10.0f, 20.0f, 10.0f));
        rootNode.addLight(sun2);
    }

    private void parseFile(String filePath) {
        int verticesCount = 0;
        int verticesGroupCount = 0;
        int normalsCount = 0;
        int textureCoordsCount = 0;
        int indexesCount = 0;

        File file = new File(filePath);
        BufferedReader reader = null;
        try {
            reader = new BufferedReader(new FileReader(file));
            String text;
            while ((text = reader.readLine()) != null) {
                if (text.contains("#")) {
                } else if (text.contains("vs")) {
                    verticesBody = new Vector3f[Integer.parseInt(text.split(" ")[1])];
                    verticesHead = new Vector3f[Integer.parseInt(text.split(" ")[1])];
                    verticesLeftLeg = new Vector3f[Integer.parseInt(text.split(" ")[1])];
                    verticesRightLeg = new Vector3f[Integer.parseInt(text.split(" ")[1])];
                    verticesLeftArm = new Vector3f[Integer.parseInt(text.split(" ")[1])];
                    verticesRightArm = new Vector3f[Integer.parseInt(text.split(" ")[1])];
                } else if (text.contains("ns")) {
                    normals = new float[3 * Integer.parseInt(text.split(" ")[1])];
                } else if (text.contains("tcs")) {
                    texCoords = new Vector2f[Integer.parseInt(text.split(" ")[1])];
                } else if (text.contains("is")) {
                    indexes = new int[Integer.parseInt(text.split(" ")[1])];
                } else if (text.contains("v")) {
                    if (verticesGroupCount < 4) {
                        verticesBody[verticesCount] = new Vector3f(
                                Float.parseFloat(text.substring(2).split(", ")[0]),
                                Float.parseFloat(text.substring(2).split(", ")[1]),
                                Float.parseFloat(text.substring(2).split(", ")[2]));
                    } else if (verticesGroupCount < 8) {
                        // TODO: (Sub-Step 1.6) ADD YOUR CODE HERE!
                        
                    } else if (verticesGroupCount < 12) {
                        // TODO: (Sub-Step 1.7) ADD YOUR CODE HERE!
                        
                    } else if (verticesGroupCount < 16) {
                        // TODO: (Sub-Step 1.8) ADD YOUR CODE HERE!
                        
                    } else if (verticesGroupCount < 20) {
                        // TODO: (Sub-Step 1.9) ADD YOUR CODE HERE!
                        
                    } else if (verticesGroupCount < 24) {
                        // TODO: (Sub-Step 1.10) ADD YOUR CODE HERE!
                        
                    }
                    verticesGroupCount++;
                    verticesCount++;
                    if (verticesCount % 4 == 0) {
                        verticesCount = 0;
                    }
                } else if (text.contains("n")) {
                    normals[normalsCount] = Float.parseFloat(text.substring(2).split(", ")[0]);
                    normals[normalsCount + 1] = Float.parseFloat(text.substring(2).split(", ")[1]);
                    normals[normalsCount + 2] = Float.parseFloat(text.substring(2).split(", ")[2]);
                    normalsCount += 3;
                } else if (text.contains("tc")) {
                    texCoords[textureCoordsCount] = new Vector2f(
                            Float.parseFloat(text.substring(3).split(", ")[0]),
                            Float.parseFloat(text.substring(3).split(", ")[1]));
                    textureCoordsCount++;
                } else if (text.contains("i")) {
                    indexes[indexesCount] = Integer.parseInt(text.substring(2).split(", ")[0]);
                    indexes[indexesCount + 1] = Integer.parseInt(text.substring(2).split(", ")[1]);
                    indexes[indexesCount + 2] = Integer.parseInt(text.substring(2).split(", ")[2]);
                    indexesCount += 3;
                }
            }
        } catch (FileNotFoundException e) {
        } catch (IOException e) {
        } finally {
            try {
                if (reader != null) {
                    reader.close();
                }
            } catch (IOException e) {
            }
        }
    }

    private void initKeys() {
        inputManager.addMapping("Up", new KeyTrigger(KeyInput.KEY_T));
        inputManager.addListener(analogListener, "Up");

        /**
         * BEGIN OF EXERCISE 3 (Step 3): complete the keys initialization code
         * in the following TODO's for Step 3.
         */
        // TODO: (Sub-Step 3.1) ADD YOUR CODE HERE!
        

        // END OF EXERCISE 3 (Step 3). ----------------------------------------- 
    }
    private AnalogListener analogListener = new AnalogListener() {
        @Override
        public void onAnalog(String name, float value, float tpf) {
            Vector3f oldRobotTranslation = new Vector3f(robot.getLocalTranslation());
            Vector3f oldBox1NodeTranslation = new Vector3f(box1Node.getLocalTranslation());

            if (name.equals("Up")) {
                robot.setLocalTranslation(robot.getLocalTranslation().add(new Vector3f(0, .1f, 0)));
                box1Node.setLocalTranslation(box1Node.getLocalTranslation().add(new Vector3f(0, .1f, 0)));
            }

            /**
             * BEGIN OF EXERCISE 4 (Step 4): complete the keys behavior code in
             * the following TODO's for Step 4.
             */
            // TODO: (Sub-Step 4.1) ADD YOUR CODE HERE!
            

            // END OF EXERCISE 4 (Step 4). ------------------------------------- 

            /**
             * BEGIN OF EXERCISE 5 (Step 5): implement the GJK algorithm code in
             * the following TODO's for Step 5.
             */
            if (GJK(box1Node, box2Node)) {
                robot.setLocalTranslation(oldRobotTranslation);
                box1Node.setLocalTranslation(oldBox1NodeTranslation);
            }

            // END OF EXERCISE 5 (Step 5). ------------------------------------- 
        }
    };

    //**************************************************************************
    // GJK code.
    //**************************************************************************
    /**
     *
     * @param AABB1
     * @param AABB2
     * @return
     */
    private boolean GJK(Node AABB1, Node AABB2) {

        Vector3f v11, v12, v13, v14, v21, v22, v23, v24;

        v11 = new Vector3f(min1.getX(), min1.getY(), 0).add(AABB1.getWorldBound().getCenter());
        v12 = new Vector3f(min1.getX(), max1.getY(), 0).add(AABB1.getWorldBound().getCenter());
        v13 = new Vector3f(max1.getX(), min1.getY(), 0).add(AABB1.getWorldBound().getCenter());
        v14 = new Vector3f(max1.getX(), max1.getY(), 0).add(AABB1.getWorldBound().getCenter());

        ArrayList<Vector3f> shape1 = new ArrayList<>();
        shape1.add(v11);
        shape1.add(v12);
        shape1.add(v14);

        ArrayList<Vector3f> shape2 = new ArrayList<>();
        shape2.add(v11);
        shape2.add(v13);
        shape2.add(v14);

        v21 = new Vector3f(min2.getX(), min2.getY(), 8);
        v22 = new Vector3f(min2.getX(), max2.getY(), 8);
        v23 = new Vector3f(max2.getX(), min2.getY(), 8);
        v24 = new Vector3f(max2.getX(), max2.getY(), 8);

        ArrayList<Vector3f> shape3 = new ArrayList<>();
        shape3.add(v21);
        shape3.add(v22);
        shape3.add(v24);

        ArrayList<Vector3f> shape4 = new ArrayList<>();
        shape4.add(v21);
        shape4.add(v23);
        shape4.add(v24);

        return BodiesIntersect(shape1, shape3)
                || BodiesIntersect(shape1, shape4)
                || BodiesIntersect(shape2, shape3)
                || BodiesIntersect(shape2, shape4);
    }

    /**
     * Given the vertices (in any order) of two convex 3D bodies, calculates
     * whether they intersect.
     *
     * @param shape1
     * @param shape2
     * @return TRUE if two convex 3D bodies intersect, FALSE otherwise
     */
    private boolean BodiesIntersect(ArrayList<Vector3f> shape1, ArrayList<Vector3f> shape2) {
        // TODO: (Sub-Step 5.1) ADD YOUR CODE HERE!
        //For initial point, just take the difference between any two vertices 
        //(in this case - the first ones of each shape).
        

        //Choose a search direction S, the negate of MaxPointInMinkDiffAlongDir 
        //from shape1 to shape2 in the direction of initialPoint.
        

        //Get the first Minkowski Difference point, i.e., initialize a 
        //ArrayList<Vector3f> simplex with S.
        

        for (int i = 0; i < MaxIterations; i++) {
            //Get our next simplex point A toward the origin, 
            //MaxPointInMinkDiffAlongDir from shape1 to shape2 in the direction 
            //of Direction.
            

            //If we move toward the origin and didn't pass it then we never will 
            //and there's no intersection, if the dot product of A with 
            //Direction is negative return false. 
            

            //Otherwise we add the new point A to the simplex and process it.
            

            //Here we either find a collision or we find the closest feature of
            //the simplex to the origin, make that the new simplex and update 
            //the direction to move toward the origin from that feature, i.e., 
            //if updating the simplex and its direction true, return true.
            
        }

        return false;
    }

    /**
     * Updates the current simplex and the direction in which to look for the
     * origin. Called DoSimplex in the video lecture.
     *
     * @param simplex A list of points in the current simplex. The last point in
     * the list must be the last point added to the simplex.
     * @param direction
     * @return TRUE if an intersection found on this iteration, FALSE otherwise.
     */
    private boolean UpdateSimplexAndDirection(ArrayList<Vector3f> simplex) {
        //if the simplex is a line
        if (simplex.size() == 2) { 
            //A is the point added last to the simplex
            //B is the first point of the simplex
            //AB = B-A
            //AO = -A
            Vector3f A = simplex.get(1);
            Vector3f B = simplex.get(0);
            Vector3f AB = B.subtract(A);
            Vector3f AO = A.negate();

            System.out.println("simplex2: " + Direction);
            //Direction equals AO except if the dot product of A with AO greater 
            //than 0. Otherwise, Direction equals triple produtc of AB with AO.
            if (AB.dot(AO) > 0) {
                Direction = AB.cross(AO).cross(AB);
            } else {
                Direction = AO;
            } //end block
            System.out.println("simplex2: " + Direction);
        } //if the simplex is a triangle
        else if (simplex.size() == 3) {
            // TODO: (Sub-Step 5.2) ADD YOUR CODE HERE!   
            //A is the point added last to the simplex
            //B is the second point added last to the simplex
            //C is the first point of the simplex
            //AO = -A
            //AB = B-A
            //AC = C-A
            //ABC = cross product of AB with AC

            
            //If dot product of the result, of cross product of ABC with AC, with 
            //AO greater than 0
                //If dot product of AC with AO greater than 0
                    //clear the simplex and add to the simplex C and A.                                


                    //Direction equals triple cross product of AC with AO.

                //else if dot product of AB with AO greater than 0 
                    //clear the simplex and add to the simplex B and A.


                    //Direction equals triple cross product of AB with AO.

                //else 
                    //clear the simplex and add to the simplex A


                    //Direction = A0

                //end block
            //else  
                //If dot product of the result, of cross product of AB with ABC, with 
                //AO greater than 0
                    //If dot product of AB with AO greater than 0
                        //clear the simplex and add to the simplex B and A.


                        //Direction equals triple cross product of AB with AO.

                    //else 
                        //clear the simplex and add to the simplex A


                        //Direction = A0

                    //end block
                //else                 
                    //If dot product of ABC with AO greater than 0
                        //the simplex stays A, B, C, i.e., Direction = ABC

                    //else
                        //clear the simplex and add to the simplex B, C, and A.


                        //Direction equals the negation of ABC 
                    
                    //end block
                //end block
            //end block
        } //if the simplex is a tetrahedron
        else //if (simplex.size() == 4)
        {       
            //A is the point added last to the simplex
            

            //the side (positive or negative) of B, C and D relative to the planes of ACD, ADB and ABC respectively
            

            //whether the origin is on the same side of ACD/ADB/ABC as B, C and D respectively
            
            
            //if the origin is on the same side as all B, C and D, the origin is inside the tetrahedron and thus there is a collision            
                
            //if the origin is not on the side of B relative to ACD
                //B is farthest from the origin among all of the tetrahedron's points, so remove it from the list and go on with the triangle case
                
                
                //the new direction is on the other side of ACD, relative to B
                

            //if the origin is not on the side of C relative to ADB
                //C is farthest from the origin among all of the tetrahedron's points, so remove it from the list and go on with the triangle case
                
                
                //the new direction is on the other side of ADB, relative to C                
            
            //if the origin is not on the side of D relative to ABC            
                //D is farthest from the origin among all of the tetrahedron's points, so remove it from the list and go on with the triangle case
                
                //the new direction is on the other side of ABC, relative to D                
                
            //end block

            //go on with the triangle case
            //Maybe we should restrict the depth of the recursion, 
            //just like we restricted the number of iterations in BodiesIntersect?
            
        } //end block

        //no intersection found on this iteration
        return false;
    }

    /**
     * Finds the farthest point along a given direction of the Minkowski
     * difference of two convex polyhedra. Called Support in the video lecture:
     * max(D.Ai) - max(-D.Bj)
     *
     * @param shape1
     * @param shape2
     * @param direction
     * @return the farthest point along a given direction of the Minkowski
     * difference of two convex polyhedra.
     */
    private Vector3f MaxPointInMinkDiffAlongDir(
            ArrayList<Vector3f> shape1, ArrayList<Vector3f> shape2, Vector3f direction) {
        Vector3f max = new Vector3f();

        // TODO: (Sub-Step 5.3) ADD YOUR CODE HERE!           
        

        return max;
    }

    /**
     * Finds the farthest point along a given direction of a convex polyhedron.
     *
     * @param shape
     * @param direction
     * @return the farthest point along a given direction of a convex
     * polyhedron.
     */
    private static Vector3f MaxPointAlongDirection(ArrayList<Vector3f> shape, Vector3f direction) {
        Vector3f max = new Vector3f(shape.get(0));

        // TODO: (Sub-Step 5.4) ADD YOUR CODE HERE! 
        //for each shape vertice 
        //if the distance along a given direction of max greater than the 
        //distance along a given direction of v, max equals v 
        

        return max;
    }
}
