package mygame;

import com.jme3.app.SimpleApplication;
import com.jme3.bullet.BulletAppState;
import com.jme3.bullet.collision.shapes.SphereCollisionShape;
import com.jme3.bullet.collision.shapes.CollisionShape;
import com.jme3.bullet.control.CharacterControl;
import com.jme3.bullet.control.RigidBodyControl;
import com.jme3.bullet.control.VehicleControl;
import com.jme3.bullet.util.CollisionShapeFactory;
import com.jme3.input.KeyInput;
import com.jme3.input.controls.ActionListener;
import com.jme3.input.controls.KeyTrigger;
import com.jme3.light.AmbientLight;
import com.jme3.light.DirectionalLight;
import com.jme3.material.Material;
import com.jme3.math.ColorRGBA;
import com.jme3.math.FastMath;
import com.jme3.math.Quaternion;
import com.jme3.math.Vector3f;
import com.jme3.scene.Geometry;
import com.jme3.scene.Node;
import com.jme3.scene.Spatial;
import com.jme3.scene.shape.Box;

/**
 * Every JME3 game is an instance of the com.jme3.app.SimpleApplication class.
 * The SimpleApplication class is the simplest example of an application: It
 * manages a 3D scene graph, checks for user input, updates the game state, and
 * automatically draws the scene to the screen. These are the core features of a
 * game engine. You extend this simple application and customize it to create
 * your game.
 *
 * @author G. Amador \& A. Gomes
 */
public class Main extends SimpleApplication
        implements ActionListener {

    private BulletAppState bulletAppState;
    private VehicleControl carObstacle;
    private RigidBodyControl cityObstacle,
            upWallObstacle,
            downWallObstacle,
            eastWallObstacle,
            westWallObstacle,
            northWallObstacle,
            southWallObstacle;
    private CharacterControl player;
    private Vector3f walkDirection = new Vector3f();
    private boolean left = false, right = false, up = false, down = false;
    //Temporary vectors used on each frame.
    //They here to avoid instanciating new vectors on each frame
    private Vector3f camDir = new Vector3f();
    private Vector3f camLeft = new Vector3f();
    private Node carNode;

    /*
     * You start every JME3 game from the main() method, as every standard Java
     * application: 
     * 1 - Instantiate your SimpleApplication-based class. 
     * 2 - The application's start() method to start the game engine.
     *
     * @param args
     */
    public static void main(String[] args) {
        Main app = new Main();            // instantiate the game
        //app.setDisplayFps(false);       // TODO: UNCOMMENT TO HIDE FPS INFO ON APPLICATION RUN
        //app.setDisplayStatView(false);  // TODO: UNCOMMENT TO HIDE STATS VIEW INFO ON APPLICATION RUN
        //app.setShowSettings(false);     // TODO: UNCOMMENT TO HIDE SETTINGS MENU ON STARTUP
        app.start();                      // start the game
    }

    @Override
    public void simpleInitApp() {
        // We re-use the flyby camera for rotation, while positioning is handled by physics
        viewPort.setBackgroundColor(new ColorRGBA(0.7f, 0.8f, 1f, 1f));
        flyCam.setMoveSpeed(10);
        setUpKeys();
        setUpLight();

        /**
         * Set up Physics
         */
        bulletAppState = new BulletAppState();
        // Since bullet is not (yet) multithreaded or GPU accelerated, the jME3
        // implementation allows to run each physics space on a separate thread
        // that is executed in parallel to rendering.
        //bulletAppState.setThreadingType(BulletAppState.ThreadingType.PARALLEL);   // TODO: UNCOMMENT TO RUN PARALLEL JBULLET 
        stateManager.attach(bulletAppState);
        bulletAppState.getPhysicsSpace().enableDebug(assetManager);               // TODO: UNCOMMENT TO ENABLE JBULLET DEBUG

        /* create a SkyBox of dimension (100, 100, 100) */
        /**
         * BEGIN OF EXERCISE 10 (Step 10): complete the SkyBox wall code in the
         * following TODO's for Step 10.
         */
        // north wall
        Material northWallMat = new Material(assetManager,
                "Common/MatDefs/Misc/Unshaded.j3md");                   // set the north wall's material generic properties
        northWallMat.setTexture("ColorMap", assetManager.loadTexture(
                "Textures/Skybox/SkyBox2_north.jpg"));                  // associate a single texture with the north wall material
        Box northWallBox = new Box(100, 100, 0);                        // create cube shape of size (100,100,0)
        Geometry northWall = new Geometry("northWall", northWallBox);   // create the north wall cube geometry from the shape    
        northWall.setLocalTranslation(0, 100, -100);                    // translate north wall to location (0, 100, -100)   
        //northWall.setLocalTranslation(new Vector3f(0, 100, -100));      // translate north wall to location (0, 100, -100)   
        northWall.setMaterial(northWallMat);                            // set the north wall's material

        // south wall
        // TODO: (Sub-Step 10.1) ADD YOUR CODE HERE!
        

        // east wall    
        // TODO: (Sub-Step 10.2) ADD YOUR CODE HERE!
        

        // west wall
        // TODO: (Sub-Step 10.3) ADD YOUR CODE HERE!
        

        // This quaternion stores a 270 degree rolling rotation around the Y axis
        Quaternion roll270 = new Quaternion();
        roll270.fromAngleAxis(FastMath.PI * 3 / 2, new Vector3f(0, 1, 0));

        // up wall
        // TODO: (Sub-Step 10.4) ADD YOUR CODE HERE!
        

        // down wall        
        Material downWallMat = new Material(assetManager,
                "Common/MatDefs/Misc/Unshaded.j3md");
        downWallMat.setTexture("ColorMap", assetManager.loadTexture(
                "Textures/Skybox/SkyBox2_down.jpg"));
        Box downWallBox = new Box(100, 0, 100);
        Geometry downWall = new Geometry("downWall", downWallBox);
        downWall.setLocalRotation(roll270);
        downWall.setLocalTranslation(new Vector3f(0, 0, 0));
        downWall.setMaterial(downWallMat);

        // END OF EXERCISE 10 (Step 10). ---------------------------------------   

        /**
         * BEGIN OF EXERCISE 11 (Step 11): complete code to setup car and city
         * nodes in the following TODO's for Step 11.
         */
        // This quaternion stores a 180 degree rolling rotation around the Y axis
        Quaternion roll180 = new Quaternion();
        roll180.fromAngleAxis(FastMath.PI, new Vector3f(0, 1, 0));

        // creat red car model node whose geometry is located at 
        // "Models/cars/carRed/Car.obj", with rotation roll180, 
        // translation (0, 1, 0.15) and scale factor of 0.35   
        Spatial carRedGeom = assetManager.loadModel(
                "Models/cars/carRed/Car.obj");
        carNode = new Node("carNode");
        carNode.attachChild(carRedGeom);
        carNode.setLocalRotation(roll180);
        carNode.setLocalTranslation(0, 1, 0.15f);   // translate the model 1 in the Y axis 0.15 in the Z axis 
        carNode.scale(0.35f);                       // reduce the model to near 1/3 of its size

        // creat city model node whose geometry is located at 
        // "Models/city/city.obj", with a scale factor of 5
        // TODO: (Sub-Step 11.1) ADD YOUR CODE HERE!
        

        // END OF EXERCISE 11 (Step 11). ---------------------------------------        

        /**
         * BEGIN OF EXERCISE 12 (Step 12): complete code to set up collision
         * detection for the walls, the city, and the car and attach the walls,
         * the city, and the car to the rootnode and the physics space in the
         * following TODO's for Step 12.
         */
        // Set up collision detection for the north wall by creating a
        // collision shape and a static RigidBodyControl with mass zero.
        CollisionShape northWallShape =
                CollisionShapeFactory.createMeshShape((Geometry) northWall);
        northWallObstacle = new RigidBodyControl(northWallShape, 0);
        northWall.addControl(northWallObstacle);

        // Attach the north wall to the rootnode and the physics space,
        // to make it appear in the game world.
        rootNode.attachChild(northWall);
        bulletAppState.getPhysicsSpace().add(northWallObstacle);

        // south wall
        // TODO: (Sub-Step 12.1) ADD YOUR CODE HERE!
        

        // east wall    
        // TODO: (Sub-Step 12.2) ADD YOUR CODE HERE!
        

        // west wall
        // TODO: (Sub-Step 12.3) ADD YOUR CODE HERE!
        

        // up wall
        // TODO: (Sub-Step 12.4) ADD YOUR CODE HERE!
        

        // We set up collision detection for the down wall by creating a
        // collision shape and a static RigidBodyControl with mass zero.
        CollisionShape downWallShape =
                CollisionShapeFactory.createMeshShape((Geometry) downWall);
        downWallObstacle = new RigidBodyControl(downWallShape, 0);
        downWall.addControl(downWallObstacle);

        // Attach the down wall to the rootnode and the physics space,
        // to make it appear in the game world.
        rootNode.attachChild(downWall);
        bulletAppState.getPhysicsSpace().add(downWallObstacle);

        // red car
        //CollisionShape carShape =
        //        CollisionShapeFactory.createMeshShape((Node) carNode);
        //carObstacle = new VehicleControl(carShape, 0);
        //carNode.addControl(carObstacle);
        rootNode.attachChild(carNode);
        //bulletAppState.getPhysicsSpace().add(carObstacle);

        // city
        // TODO: (Sub-Step 12.5) ADD YOUR CODE HERE!
        

        // END OF EXERCISE 12 (Step 12). ---------------------------------------

        // We set up collision detection for the player by creating
        // a capsule collision shape and a CharacterControl.
        // The CharacterControl offers extra settings for
        // size, stepheight, jumping, falling, and gravity.
        // We also put the player in its starting position.
        SphereCollisionShape playerCollisionShape = new SphereCollisionShape(2.0f);
        player = new CharacterControl(playerCollisionShape, 0.05f);
        //player.setJumpSpeed(20);  // DO NOT UNCOMMENT! Set JumpSpeed (up) velocity value
        //player.setFallSpeed(30);  // DO NOT UNCOMMENT! Set FallSpeed (Down) velocity value
        //player.setGravity(30);    // DO NOT UNCOMMENT! Set Gravity (Down) accelaration value.
        //player.setPhysicsLocation(new Vector3f(0, 5, 0));
        player.setPhysicsLocation(carNode.getLocalTranslation());

        // Attach the player to the physics space.
        bulletAppState.getPhysicsSpace().add(player);
    }

    private void setUpLight() {
        // We add light so we see the scene
        AmbientLight al = new AmbientLight();
        al.setColor(ColorRGBA.White.mult(1.3f));
        rootNode.addLight(al);

        DirectionalLight dl = new DirectionalLight();
        dl.setColor(ColorRGBA.White);
        dl.setDirection(new Vector3f(2.8f, -2.8f, -2.8f).normalizeLocal());
        rootNode.addLight(dl);
    }

    /**
     * We over-write some navigational key mappings here, so we can add
     * physics-controlled walking:
     */
    private void setUpKeys() {
        inputManager.addMapping("Left", new KeyTrigger(KeyInput.KEY_A));
        inputManager.addMapping("Right", new KeyTrigger(KeyInput.KEY_D));
        inputManager.addMapping("Up", new KeyTrigger(KeyInput.KEY_W));
        inputManager.addMapping("Down", new KeyTrigger(KeyInput.KEY_S));
        inputManager.addListener(this, "Left");
        inputManager.addListener(this, "Right");
        inputManager.addListener(this, "Up");
        inputManager.addListener(this, "Down");
    }

    /**
     * These are our custom actions triggered by key presses. We do not walk
     * yet, we just keep track of the direction the user pressed.
     */
    @Override
    public void onAction(String binding, boolean isPressed, float tpf) {
        switch (binding) {
            case "Left":
                left = isPressed;
                break;
            case "Right":
                right = isPressed;
                break;
            case "Up":
                up = isPressed;
                break;
            case "Down":
                down = isPressed;
                break;
        }
    }

    /**
     * This is the main event loop--walking happens here. We check in which
     * direction the player is walking by interpreting the camera direction
     * forward (camDir) and to the side (camLeft). The setWalkDirection()
     * command is what lets a physics-controlled player walk. We also make sure
     * here that the camera moves with player.
     */
    @Override
    public void simpleUpdate(float tpf) {
        camDir.set(cam.getDirection()).multLocal(0.6f);
        camLeft.set(cam.getLeft()).multLocal(0.4f);
        //Vector3f camDirPast = new Vector3f(camDir);
        //Vector3f camLeftPast = new Vector3f(camLeft);
        // Vector3f playerPastLocation = new Vector3f(player.getPhysicsLocation());
        walkDirection.set(0, 0, 0);
        if (left) {
            walkDirection.addLocal(camLeft);
        }
        if (right) {
            walkDirection.addLocal(camLeft.negate());
        }
        if (up) {
            walkDirection.addLocal(camDir);
        }
        if (down) {
            walkDirection.addLocal(camDir.negate());
        }
        player.setWalkDirection(walkDirection);

        // Set camera translation to equal player translation
        cam.setLocation(new Vector3f(
                player.getPhysicsLocation().x,
                10, // Set camera Y position
                player.getPhysicsLocation().z));

        // Set car rotation to equal the camera rotation around the Y axis 
        carNode.setLocalRotation(new Quaternion(0, cam.getRotation().getY(), 0, cam.getRotation().getW()));

        // Set car translation to equal player translation
        carNode.setLocalTranslation(
                player.getPhysicsLocation().x,
                carNode.getLocalTranslation().y, // Y position stays the same
                player.getPhysicsLocation().z);
    }
}
